/****************************************************************************/
/*                         SQL Server 2016 Virtual Forum                    */
/*                                                                          */
/*                  Written by Dmitri V. Korotkevitch                       */
/*                      http://aboutsqlserver.com                           */
/*                        dk@aboutsqlserver.com                             */
/****************************************************************************/
/*                    Tiered Storage: Data Movement                         */
/****************************************************************************/
set nocount on
set xact_abort on
go

/* This script requires SQL Server 2016. In SQL Server 2014 you can replace triggers
with the code on data access tier */

use InMemoryOLTP2016
go

if exists(select * from sys.triggers  where name = 'trgAfterUpdate') drop trigger dbo.trgAfterUpdate;
if exists(select * from sys.triggers  where name = 'trgAfterDelete') drop trigger dbo.trgAfterDelete;

if exists(select * from sys.tables t join sys.schemas s on t.schema_id = s.schema_id where s.name = 'dbo' and t.name = 'PaymentsUpdateQueue') drop table dbo.PaymentsUpdateQueue;
if exists(select * from sys.tables t join sys.schemas s on t.schema_id = s.schema_id where s.name = 'dbo' and t.name = 'PaymentsDeleteQueue') drop table dbo.PaymentsDeleteQueue;
if exists(select * from sys.tables t join sys.schemas s on t.schema_id = s.schema_id where s.name = 'dbo' and t.name = 'Payments2016_05_Tmp') drop table dbo.Payments2016_05_Tmp;
go

create table dbo.PaymentsUpdateQueue
(
	ID int not null identity(1,1)
		constraint PK_PaymentsUpdateQueue
		primary key nonclustered hash
		with (bucket_count=262144),
	PaymentId uniqueidentifier not null, 
)
with (memory_optimized=on, durability=schema_and_data)
go


create table dbo.PaymentsDeleteQueue
(
	ID int not null identity(1,1)
		constraint PK_PaymentsDeleteQueue
		primary key nonclustered hash
		with (bucket_count=262144),
	PaymentId uniqueidentifier not null
)
with (memory_optimized=on, durability=schema_and_data)
go

create trigger trgAfterUpdate on dbo.Payments2016_05
with native_compilation, schemabinding
after update
as
begin atomic with
(
    transaction isolation level = snapshot
    ,language = N'English'
)
	insert into dbo.PaymentsUpdateQueue(PaymentId)
		select PaymentId 
		from inserted
end
go

create trigger trgAfterDelete on dbo.Payments2016_05
with native_compilation, schemabinding
after delete
as
begin atomic with
(
    transaction isolation level = snapshot
    ,language = N'English'
)
	insert into dbo.PaymentsDeleteQueue(PaymentId)
		select PaymentId 
		from deleted
end
go

create table dbo.Payments2016_05_Tmp
(
	PaymentDate datetime2(0) not null,
	PaymentId uniqueidentifier not null, 
	CustomerId int not null,
	Amount money not null,
	Status tinyint not null,

	check (PaymentDate >= '2016-05-01' and PaymentDate < '2016-06-01')
)
on [FG2016]
go

create unique clustered index IDX_Payments2016_05_Tmp_PaymentDate_PaymentId
on dbo.Payments2016_05_Tmp(PaymentDate, PaymentId)
with (data_compression=row)
on [FG2016]
go

create nonclustered index IDX_Payments2016_05_Tmp_CustomerId
on  dbo.Payments2016_05_Tmp(CustomerId)
with (data_compression=row)
on [FG2016]
go

create nonclustered index IDX_Payments2016_05_Tmp_PaymentId
on  dbo.Payments2016_05_Tmp(PaymentId)
with (data_compression=row)
on [FG2016]
go

-- Step 1: Move data from the main table
insert into dbo.Payments2016_05_Tmp(PaymentDate, PaymentId, CustomerId, Amount, Status)
	select PaymentDate, PaymentId, CustomerId, Amount, Status
	from dbo.Payments2016_05 with (snapshot)
go

-- Step 2: Update/Delete
select count(*) as [In Update Queue] from dbo.PaymentsUpdateQueue with (snapshot)
select count(*) as [In Delete Queue] from dbo.PaymentsDeleteQueue with (snapshot)
go

declare
	@MaxUpdateId int
	,@MaxDeleteId int

select @MaxUpdateId = max(ID) from dbo.PaymentsUpdateQueue with (snapshot);
select @MaxDeleteId = max(ID) from dbo.PaymentsDeleteQueue with (snapshot);

begin tran
	 if @MaxUpdateId is not null
	begin
		update t
		set t.Amount = s.Amount, t.Status = s.Status
		from 
			dbo.PaymentsUpdateQueue q with (snapshot) join dbo.Payments2016_05 s with (snapshot) on
				q.PaymentId = s.PaymentId 
			join dbo.Payments2016_05_Tmp t on
				t.PaymentId = s.PaymentId
		where
			q.ID <= @MaxUpdateId;
		
		delete from dbo.PaymentsUpdateQueue with (snapshot)
		where ID <= @MaxUpdateId;
	end;

	if @MaxDeleteId is not null
	begin
		delete from t
		from 
			dbo.PaymentsDeleteQueue q with (snapshot) join dbo.Payments2016_05_Tmp t on
				t.PaymentId = q.PaymentId
		where
			q.ID <= @MaxDeleteId;
		
		delete from dbo.PaymentsDeleteQueue with (snapshot)
		where ID <= @MaxDeleteId;
	end

	select count(*) as [In Update Queue] from dbo.PaymentsUpdateQueue with (snapshot)
	select count(*) as [In Delete Queue] from dbo.PaymentsDeleteQueue with (snapshot)
commit
go

-- Step 3: Replace
set xact_abort on
set deadlock_priority 10
go

begin tran -- Keeping locks in place
go

-- Obtaining Sch-M locks on SPs
alter proc dbo.UpdatePayments2016
(
	@PaymentDate datetime2(0) 
	,@PaymentId uniqueidentifier
	,@Amount money
	,@Status tinyint
)
as
	update dbo.Payments2016
	set Amount = @Amount, @Status = @Status
	where PaymentId = @PaymentId and @PaymentDate = @PaymentDate
go

alter proc dbo.DeletePayments2016
(
	@PaymentDate datetime2(0) 
	,@PaymentId uniqueidentifier
)
as
	delete from dbo.Payments2016
	where PaymentId = @PaymentId and @PaymentDate = @PaymentDate
go

update t
set t.Amount = s.Amount, t.Status = s.Status
from 
	dbo.PaymentsUpdateQueue q with (snapshot) join dbo.Payments2016_05 s with (snapshot) on
		q.PaymentId = s.PaymentId 
	join dbo.Payments2016_05_Tmp t on
		t.PaymentId = s.PaymentId;

delete from t
from 
	dbo.PaymentsDeleteQueue q with (snapshot) join dbo.Payments2016_05_Tmp t on
		t.PaymentId = q.PaymentId;

alter table dbo.Payments2016
	drop constraint CHK_Payment2016_01_05
go

alter table dbo.Payments2016_05_Tmp switch to dbo.Payments2016 partition 5
go

alter view dbo.Payments(PaymentDate, PaymentId, CustomerId, Amount, Status)
as
	select PaymentDate, PaymentId, CustomerId, Amount, Status
	from dbo.Payments2016_06
	
	union all

	select PaymentDate, PaymentId, CustomerId, Amount, Status
	from dbo.Payments2016

	union all

	select PaymentDate, PaymentId, CustomerId, Amount, Status
	from dbo.Payments2015
go
commit
go

drop table dbo.Payments2016_05
go

select max(PaymentDate) from dbo.Payments2016
go

-- Housekeeping
alter index IDX_Payments2016_PaymentDate_PaymentId on dbo.Payments2016 rebuild partition = 5
with (online=on)
go

alter index IDX_Payments2016_CustomerId on dbo.Payments2016 rebuild partition = 5
with (online=on)
go

alter index IDX_Payments2016_PaymentId on dbo.Payments2016 rebuild partition = 5
with (online=on)
go